{
module LangParse where

import LangLex
import CST
import Control.Monad.Error
import Debug.Trace
}

%name parseProg Program

%monad { Either String } { (>>=) } { return }
%tokentype { LangLex.PToken }

%token '='     { (_,TkEq) }
       "()"    { (_,TkUnit) }
       '('     { (_,TkLPar) }
       ')'     { (_,TkRPar) }
       '{'     { (_,TkLBrace) }
       '}'     { (_,TkRBrace) }
       '['     { (_,TkLBracket) }
       ']'     { (_,TkRBracket) }
       '|'     { (_,TkPar) }
       "::"    { (_,TkColons) }
       ':'     { (_,TkColon) }
       ';'     { (_,TkSemiColon) }
       "->"    { (_,TkArr) }
       '<'     { (_,TkLABracket ) }
       '>'     { (_,TkRABracket ) }
       "where" { (_,TkKey "where") }
       "atomic" { (_,TkKey "atomic") }
       "oneway" { (_,TkKey "oneway") }
       "inversepair" { (_,TkKey "inversepair") }
       "with" { (_,TkKey "with") }
       "data" { (_,TkKey "data") }
       "abort" { (_,TkKey "abort") }
       "++"    { (_,TkAppend ) }
       '+'     { (_,TkPlus ) }
       '-'     { (_,TkMinus ) }
       '*'     { (_,TkAst ) }
       '/'     { (_,TkSlash ) }
       '#'     { (_,TkSharp ) }
       '$'     { (_,TkDollar) }
       ','     { (_,TkComma) }
       cat     { (_,TkDot) }
       '~'     { (_,TkTilde) }
       '_'     { (_,TkUnderscore) }
       '@'     { (_,TkAt) }
       string  { (_,TkString $$) }
       integer { (_,TkInt $$) }
       sident  { (_,TkSIdent $$) }             
       ident   { (_,TkIdent $$) }
       cident  { (_,TkCIdent $$) }
       bident  { (_,TkBIdent $$) }


%%


Program 
: Decl ';' Program { $1 : $3 }
| ';' Program      { $2 }
|                  { [] }




Decl
: "data" cident '=' TE      { TDecl $2 $4 }
| ident '(' Pats ')' '=' E  { FDecl $1 $3 $6 }


TE
: TE '|'  TE1    { TSel $1 $3 }
| TE1            { $1 }

TE1
: TE2 cat TE1       { TApp $1 $3 }
| TE2               { $1 }

TE2
: bident TRE '(' TE ')'  { TCon $1 $2 $4 }
| bident TRE '(' ')'     { TCon $1 $2 TNil }
| bident TRE             { TCon $1 $2 TNil }
| TE2 '*'        { TStar $1 }
| TE2 '+'        { TApp $1 (TStar $1) }
| cident         { TFun $1 }
| string         { case $1 of 
                     [] -> TNil
                     x -> foldr1 TApp $ map TChar $1 }
| '~' TRE '(' TE ')'     { TAnyLabel $2 $4 }
| '~' TRE '(' ')'        { TAnyLabel $2 TNil }
| '~' TRE                { TAnyLabel $2 TNil }
| '_'            { TAnyChar }
| '$'            { TNil } 
| '(' TE ')'     { $2 }


TRE
: '{' TRe '}' { $2 }
|             { TREmp }

TRe 
: ident ':' TE ',' TRe { ($1,$3) `TRCons` $5 }
| ident ':' TE         { ($1,$3) `TRCons` TREmp }
|                      { TREmp }

Pats
: Pat ',' Pats  { $1 : $3 }
| Pat           { [$1] }

Pat
: Pat1 cat Pat   { PApp $1 $3 }
| Pat1           { $1 }

Pat1
: '$'                      { PNil }
| bident RPt '(' Pat ')'   { PCon $1 $2 $4 }
| bident RPt '(' ')'       { PCon $1 $2 PNil }
| bident RPt               { PCon $1 $2 PNil }
| ident "::" TE2   { PVar $1 $3 }
| ident            { PVar $1 (TFun "Any") }
| '(' Pat ')'      { $2 }
| Lit              { PLit $1 }

RPt 
: '{' RP '}'     { $2 }
|                { REmp }

RP 
: ident ':' Pat1 ',' RP { ($1,$3) `RCons` $5 }
| ident ':' Pat1        { ($1,$3) `RCons` REmp  }
| sident               { RVar $1 }
|                      { REmp }

-- RPv 
-- :  ident               { PRVar $1 }
-- |  Lit                 { PLit  $1 }

E
: E cat E1              { EApp $1 $3 }
| E1                    { $1 }

E1
: '$'                   { ENil }
| bident REx '(' E ')' { ECon $1 $2 $4 }
| bident REx '(' ')'   { ECon $1 $2 ENil }
| bident REx           { ECon $1 $2 ENil }
| ident '(' Args ')'   { EFun $1 $3 }
| E2                   { $1 }

E2
: ident                { EVar $1 }
| string               { ELit $1 }
| '(' E ')'            { $2 }

REx
: REx1               { $1 }
|                    { ERec [] }

REx1
: REx2                       { $1 } 
| REx2 '#' REx1              { EUni $1 $3 }

REx2
: '{' RE '}'                 { ERec $2 }
| '{' ident '(' Args ')' '}' { EFun $2 $4 }
| '{' REx1 '}'               { $2 }

Args 
: ident ',' Args { $1 : $3 }
| ident          { [$1] }

RE 
: ident ':' E ',' RE { ($1,$3) : $5 }
| ident ':' E        { [($1,$3)] }
|                    { [] }

Lit
: string { $1 }


{
happyError tokens =
    if length tokens > 0 then
        throwError $ "ParseError at " ++ (p tokens) ++ "\n"
    else
        throwError $ "ParseError at " ++ "Unknown Position"
    where p (( (n,c),  t ):_) = "Line: " ++ show n ++ ", Column: "++show c
}
