module Main where

import LangLex
import LangParse
import qualified CST as CST
import AST
import Automaton 
import Debug.Trace
import Complementation
import Control.Monad.State
import Specialize
import System.Environment
import System.Exit
import Data.List

data Flags 
    = F_TypeSpecialize
    | F_Verbose 
    | F_Usage
      deriving (Eq,Ord,Show)

showOverlapH reg (P p) =
    swOverlapH ckOverlapH ++ swEx ckOverlapH 
    where
      ckOverlapH = [ (i,evalState (checkOverlapH (SE j) (SE k) reg) ([],[]),e) | (i,j,k,e) <- appendsAll p ]
          where
            appendsAll p = concatMap (\(_,(_,_,e)) -> appends e) p
            appends (ECon _ c r e) = appends r ++ appends e
            appends (ERec _ rs)    = concatMap (appends.snd) rs
            appends (EApp i e1 e2) = (i,expId e1, expId e2,EApp i e1 e2):appends e1 ++ appends e2
            appends _              = []
      swOverlapH [] = []
      swOverlapH ((i,b,e):rs) = 
          "  E#"++show i++": "++show b ++ "\n"++swOverlapH rs
      swEx [] = ""
      swEx ((i,b,e):rs) = 
          "  E#"++show i++": "++showP e ++ "\n"++swEx rs

showOverlapV reg (P p) =
    concatMap showMatrix $ groupBy (\(_,_,g1,_,_,_,_,_) (_,_,g2,_,_,_,_,_)->g1==g2) ckOverlapV
    where
      ckOverlapV = 
          evalState (f p p) ([],[])
          where 
            f [] [] = return []
            f [] (a:x) = f p x
            f ((i,(g,xx,e)):p) ((i',(g',xx',e')):p') =
                if g == g' && i < i' then
                    do b  <- checkOverlap (SE $ expId e) (SE $ expId e') reg
                       bs <- f p ((i',(g',xx',e')):p')
                       return $ (i,i',g,xx,xx',expId e, expId e',b):bs
                else
                    f p ((i',(g',xx',e')):p')
            f _ _ = return []
      showMatrix mat = 
          funcName ++ ":\n"
          ++ replicate maxLength ' ' ++ " | " ++ showForeach ids (\i -> take (maxLength+1) (show i++cycle " ")) ++ "\n"
          ++ replicate maxLength '-' ++ "-+-" ++ replicate (length heads) '-'++"\n"
          ++ showForeach ids (\i ->
                take (maxLength) (show i++cycle " ") ++ " | " ++ 
                     showForeach ids (\j ->
                       take (maxLength+1) (showTF (check i j)++cycle " "))++"\n")++"\n"
              where
                heads = showForeach ids (\i -> take (maxLength+1) (show i++cycle " "))
                showForeach [] f     = ""
                showForeach (x:xs) f = f x++showForeach xs f 
                check i j = case [ b || (x==y) | (x,y,_,_,_,_,_,b) <- mat, (i==x && j==y) || (i==y&&j==x) || (x==y) ] of
                              [b] -> b
                              _   -> True
                showTF True  = "O" 
                showTF False = "_"
                funcName = let (_,_,g,_,_,_,_,_) = head mat in g 
                ids = sort $ snub $ concat [ [i,j] | (i,j,_,_,_,_,_,_) <- mat ]
                maxLength = maximum $ 0:map (length.show) ids

     

      
testAST decls = 
    do 
     {-
      print $ G g
      putStrLn $ "Mutually Defined Names:"
      print $ mutually_recursive_names (G g)
      putStrLn $ "Result of Mohri & Nederhof Transformation:"
      print $ mn_transform (G g)
      putStrLn $ "Regular Approximation:"
      print $ reg
     -}
      putStrLn $ take 60 $ "== Input Program ===" ++ cycle "="
--      putStrLn $ show $ fst (cst2preast decls)
--      putStrLn $ show (G g)
      putStrLn $ show reg
      putStrLn $ showP (P p)
      putStrLn $ take 60 $ "== After Type Specialization ===" ++ cycle "="
      putStrLn $ show $ regular_approximate (G g') 
      putStrLn $ showP (P p')
--      putStrLn $ show  (G g') 
--      putStrLn $ showP (P p')
--      putStrLn $ "H Overlaps:"
--      putStrLn $ swOverlapH ckOverlapH
--      putStrLn $ "V Overlaps:"
--      putStrLn $ swOverlapV ckOverlapV
      putStrLn $ take 60 $ "== Cmpl ============" ++ cycle "="
      putStrLn $ showP $ cmpl
      putStrLn $ take 60 $ "== Inv Tupled ======" ++ cycle "="
      putStrLn $ showP $ invt
    where
--      (G g, P p) = cst2ast decls
--      reg = regular_approximate (G g) (P p)
      (G g, P p) = cst2ast decls 
      reg        = regular_approximate (G g)
      (G g',P p') = typeSpecialize reg (P p)
      (cmpl,invt) = prepare_backward (P p, reg)
      ckOverlapH = [ (i,evalState (checkOverlapH (SE j) (SE k) reg) ([],[])) | (i,j,k) <- appendsAll p ]
          where
            appendsAll p = concatMap (\(_,(_,_,e)) -> appends e) p
            appends (ECon _ c r e) = appends r ++ appends e
            appends (ERec _ rs)    = concatMap (appends.snd) rs
            appends (EApp i e1 e2) = (i,expId e1, expId e2):appends e1 ++ appends e2
            appends _              = []
      ckOverlapV = 
          evalState (f p p) ([],[])
          where 
            f [] [] = return []
            f [] (a:x) = f p x
            f ((i,(g,xx,e)):p) ((i',(g',xx',e')):p') =
                if g == g' && i < i' then
                    do b  <- checkOverlap (SE $ expId e) (SE $ expId e') reg
                       bs <- f p ((i',(g',xx',e')):p')
                       return $ (i,i',g,xx,xx',expId e, expId e',b):bs
                else
                    f p ((i',(g',xx',e')):p')
            f _ _ = return []
      swOverlapH [] = []
      swOverlapH ((i,b):rs) = 
          "Horizontal Overlaps of E#"++show i++": "++show b ++ "\n"++swOverlapH rs
      swOverlapV [] = []
      swOverlapV ((i,i',g,p,p',j,j',b):rs) = 
          show g ++ ": "++show p++ " <--> " ++ show p' ++ " ("++show i++","++show i'++")"++ ": "++show b
                                    ++ "\n" ++ swOverlapV rs

parseFlags args = f args
    where
      f [] = []
      f (x:xs) =
          case x of
            "-s" -> F_TypeSpecialize:f xs 
            "-v" -> F_Verbose:f xs
            "-h" -> F_Usage:f xs
            _ -> f xs
    
bidirectionalize flags (G org_g, P org_p) = 
    do { let org_reg = regular_approximate (G org_g) 
       ; let (reg,P p) = applyTypeSpecialize org_reg (P org_p)
       ; let (cmpl,invt) = prepare_backward (P p, reg)
       ; if F_Verbose `elem` flags then 
             do {
             ; putStr $ heading "Input Program"
             ; putStrLn $ showAutomaton org_reg (map SE (importantStates (P org_p)))
             ; putStrLn $ showP (P org_p)             
             ; if F_TypeSpecialize `elem` flags then 
                   do { putStr $ heading "After Type Specialization" 
                       ; putStrLn $ showAutomaton reg (map SE (importantStates (P p)))
                       ; putStrLn $ showP (P p) }                      
               else  
                   return ()
             ; putStr $ heading "Overlap Information"
             ; putStrLn "Horizontal Overlaps:"
             ; putStrLn $ showOverlapH reg (P p)
             ; putStrLn ""
             ; putStrLn "Vertical Overlaps:"
             ; putStrLn $ showOverlapV reg (P p) }
         else 
             return ()
       ; putStr $ heading "Complement Function"
       ; putStrLn $ showAutomaton reg (importantStatesC cmpl)
       ; putStrLn $ showP cmpl
       ; putStr $ heading "Inverted Tupled"
       ; putStrLn $ showAutomaton reg (importantStatesC invt)
       ; putStrLn $ showP invt
       ;}
    where
      applyTypeSpecialize reg (P p) 
          = if F_TypeSpecialize `elem` flags then
                let (G g', P p') = typeSpecialize reg (P p) 
                in (regular_approximate (G g'), P p')
            else
                (reg, P p)
      heading str = replicate (length str+2) '-' ++ "\n"
                    ++ " " ++ str ++ "\n"
                    ++ replicate (length str+2) '-' ++ "\n"
      showAutomaton (Automaton transes) st =
          let transes' = accessibleRules st transes 
          in show $ Automaton transes'
    


parseProgram s = -- trace (show$scanTokens s) $ 
  do {
    tokens <- scanTokens s
  ; parseProg tokens
  }

usage = do
  putStrLn "Usage:"
  putStrLn "  bidirectionalizer_f -[svh]\n"
  putStrLn $ ec "The program ``bidirectionalizer_f'' takes a forward transformation program from the stardard input, calcaltes a complement function, calculates the inverse of the tupled function of the input and the calculated complement function, and returns them."
  putStrLn "Options:"
  putStrLn "  -s  use type specialization"
  putStrLn "  -v  verbose mode"
  putStrLn "  -h  print this help message"
      where ec [] = []
            ec s  = 
                let (t,r) = splitAt 80 s
                in t ++ "\n" ++ ec r 

main = 
    do { s  <- getContents
       ; args <- getArgs
       ; flgs <- return $ parseFlags args 
       ; if F_Usage `elem` flgs then
             do { usage; exitWith ExitSuccess }
         else
             do {return ()}
       ; case parseProgram s of
           Right v -> do {
                        bidirectionalize flgs (cst2ast v)
--                      ; print v
--                      ; putStrLn "<------>"
--                      ; testAST v
                      }
           Left s -> 
                   do { putStrLn s }                      
       }